/*
 * Copyright Contributors to the fusa-gcc-plugin project
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */
#include <fstream>
#include <iomanip>
#include <iostream>

#include "cryptoutils.hpp"

using namespace std;

// https://stackoverflow.com/a/74671723
namespace crypto {
bool sha256_file_impl(string fname, vector<unsigned char> &hash) {
  unique_ptr<EVP_MD_CTX, void (*)(EVP_MD_CTX *)> evpCtx(EVP_MD_CTX_new(),
                                                        EVP_MD_CTX_free);
  EVP_DigestInit_ex(evpCtx.get(), EVP_sha256(), nullptr);

  constexpr size_t buffer_size{1 << 12};
  vector<char> buffer(buffer_size, '\0');

  ifstream fp(fname, ios::binary);
  if (!fp.is_open()) {
    cerr << "Unable to open '" << fname << "'!" << endl;
    return false;
  }
  while (fp.good()) {
    fp.read(buffer.data(), buffer_size);
    EVP_DigestUpdate(evpCtx.get(), buffer.data(), fp.gcount());
  }
  fp.close();

  hash.resize(SHA256_DIGEST_LENGTH);
  fill(hash.begin(), hash.end(), 0);
  unsigned int len;
  EVP_DigestFinal_ex(evpCtx.get(), hash.data(), &len);

  return true;
}

bool sha256_str_impl(const string &s, vector<unsigned char> &hash) {
  unique_ptr<EVP_MD_CTX, void (*)(EVP_MD_CTX *)> evpCtx(EVP_MD_CTX_new(),
                                                        EVP_MD_CTX_free);
  EVP_DigestInit_ex(evpCtx.get(), EVP_sha256(), nullptr);
  EVP_DigestUpdate(evpCtx.get(), s.c_str(), s.size());

  hash.resize(SHA256_DIGEST_LENGTH);
  fill(hash.begin(), hash.end(), 0);
  unsigned int len;
  EVP_DigestFinal_ex(evpCtx.get(), hash.data(), &len);

  return true;
}

string sha256_to_str(vector<unsigned char> &hash) {
  stringstream out;
  for (size_t i = 0; i < hash.size(); i++)
    out << setfill('0') << setw(2) << hex << int(hash[i]);
  return out.str();
}

// Compute sha256 digest of string 's' or return an empty string
string sha256_string(const string &s) {
  vector<unsigned char> hash;
  if (!sha256_str_impl(s, hash))
    return string();
  return sha256_to_str(hash);
}

// Compute sha256 digest of file 'f' or return an empty string
string sha256_file(const string &f) {
  vector<unsigned char> hash;
  if (!sha256_file_impl(f, hash))
    return string();
  return sha256_to_str(hash);
}
}; // namespace crypto
